@@ -1,17 +1,34 @@
 //! The CommonMark AST.
 
-use crate::arena_tree::Node;
+use std::borrow::Cow;
 use std::cell::RefCell;
 use std::convert::TryFrom;
 
+use crate::arena_tree;
 #[cfg(feature = "shortcodes")]
 pub use crate::parser::shortcodes::NodeShortCode;
 
-pub use crate::parser::math::NodeMath;
-pub use crate::parser::multiline_block_quote::NodeMultilineBlockQuote;
+/// Shorthand for checking if a node's value matches the given expression.
+///
+/// Note this will call `node.data()`, which will fail if the node is already
+/// mutably borrowed.
+#[macro_export]
+macro_rules! node_matches {
+    ($node:expr, $( $pat:pat )|+) => {{
+        matches!(
+            $node.data().value,
+            $( $pat )|+
+        )
+    }};
+}
 
 /// The core AST node enum.
 #[derive(Debug, Clone, PartialEq, Eq)]
+#[cfg_attr(test, derive(strum::EnumDiscriminants))]
+#[cfg_attr(
+    test,
+    strum_discriminants(vis(pub(crate)), derive(strum::VariantArray, Hash))
+)]
 pub enum NodeValue {
     /// The root of every CommonMark document.  Contains **blocks**.
     Document,
@@ -22,7 +39,7 @@ pub enum NodeValue {
     /// **Block**. A [block quote](https://github.github.com/gfm/#block-quotes).  Contains other
     /// **blocks**.
     ///
-    /// ``` md
+    /// ```markdown
     /// > A block quote.
     /// ```
     BlockQuote,
@@ -30,7 +47,7 @@ pub enum NodeValue {
     /// **Block**.  A [list](https://github.github.com/gfm/#lists).  Contains
     /// [list items](https://github.github.com/gfm/#list-items).
     ///
-    /// ``` md
+    /// ```markdown
     /// * An unordered list
     /// * Another item
     ///
@@ -48,7 +65,7 @@ pub enum NodeValue {
     ///
     /// It is required to put a blank line between terms and details.
     ///
-    /// ``` md
+    /// ```markdown
     /// Term 1
     ///
     /// : Details 1
@@ -71,7 +88,7 @@ pub enum NodeValue {
     /// **Block**. A code block; may be [fenced](https://github.github.com/gfm/#fenced-code-blocks)
     /// or [indented](https://github.github.com/gfm/#indented-code-blocks).  Contains raw text
     /// which is not parsed as Markdown, although is HTML escaped.
-    CodeBlock(NodeCodeBlock),
+    CodeBlock(Box<NodeCodeBlock>),
 
     /// **Block**. A [HTML block](https://github.github.com/gfm/#html-blocks).  Contains raw text
     /// which is neither parsed as Markdown nor HTML escaped.
@@ -95,7 +112,7 @@ pub enum NodeValue {
 
     /// **Block**. A [table](https://github.github.com/gfm/#tables-extension-) per the GFM spec.
     /// Contains table rows.
-    Table(NodeTable),
+    Table(Box<NodeTable>),
 
     /// **Block**. A table row.  The `bool` represents whether the row is the header row or not.
     /// Contains table cells.
@@ -106,11 +123,11 @@ pub enum NodeValue {
 
     /// **Inline**.  [Textual content](https://github.github.com/gfm/#textual-content).  All text
     /// in a document will be contained in a `Text` node.
-    Text(String),
+    Text(Cow<'static, str>),
 
-    /// **Inline**. [Task list item](https://github.github.com/gfm/#task-list-items-extension-).
+    /// **Block**. [Task list item](https://github.github.com/gfm/#task-list-items-extension-).
     /// The value is the symbol that was used in the brackets to mark a task item as checked, or
-    /// None if the item is unchecked.
+    /// `None` if the item is unchecked.
     TaskItem(Option<char>),
 
     /// **Inline**.  A [soft line break](https://github.github.com/gfm/#soft-line-breaks).  If
@@ -127,6 +144,10 @@ pub enum NodeValue {
     /// **Inline**.  [Raw HTML](https://github.github.com/gfm/#raw-html) contained inline.
     HtmlInline(String),
 
+    /// **Block/Inline**.  A Raw output node. This will be inserted verbatim into CommonMark and
+    /// HTML output. It can only be created programmatically, and is never parsed from input.
+    Raw(String),
+
     /// **Inline**.  [Emphasized](https://github.github.com/gfm/#emphasis-and-strong-emphasis)
     /// text.
     Emph,
@@ -143,17 +164,17 @@ pub enum NodeValue {
 
     /// **Inline**.  A [link](https://github.github.com/gfm/#links) to some URL, with possible
     /// title.
-    Link(NodeLink),
+    Link(Box<NodeLink>),
 
     /// **Inline**.  An [image](https://github.github.com/gfm/#images).
-    Image(NodeLink),
+    Image(Box<NodeLink>),
 
     /// **Inline**.  A footnote reference.
     FootnoteReference(NodeFootnoteReference),
 
     #[cfg(feature = "shortcodes")]
     /// **Inline**. An Emoji character generated from a shortcode. Enable with feature "shortcodes".
-    ShortCode(NodeShortCode),
+    ShortCode(Box<NodeShortCode>),
 
     /// **Inline**. A math span. Contains raw text which is not parsed as Markdown.
     /// Dollar math or code math
@@ -170,7 +191,7 @@ pub enum NodeValue {
     /// **Block**. A [multiline block quote](https://github.github.com/gfm/#block-quotes).  Spans multiple
     /// lines and contains other **blocks**.
     ///
-    /// ``` md
+    /// ```markdown
     /// >>>
     /// A paragraph.
     ///
@@ -181,10 +202,33 @@ pub enum NodeValue {
     MultilineBlockQuote(NodeMultilineBlockQuote),
 
     /// **Inline**.  A character that has been [escaped](https://github.github.com/gfm/#backslash-escapes)
+    ///
+    /// Enabled with [`escaped_char_spans`](crate::options::RenderBuilder::escaped_char_spans).
     Escaped,
 
     /// **Inline**.  A wikilink to some URL.
     WikiLink(NodeWikiLink),
+
+    /// **Inline**.  Underline. Enabled with `underline` option.
+    Underline,
+
+    /// **Inline**.  Subscript. Enabled with `subscript` options.
+    Subscript,
+
+    /// **Inline**.  Spoilered text.  Enabled with `spoiler` option.
+    SpoileredText,
+
+    /// **Inline**. Text surrounded by escaped markup. Enabled with `spoiler` option.
+    /// The `String` is the tag to be escaped.
+    EscapedTag(String),
+
+    /// **Block**. GitHub style alert boxes which uses a modified blockquote syntax.
+    /// Enabled with the `alerts` option.
+    Alert(Box<NodeAlert>),
+
+    /// **Block**. Block scoped subscript that acts similar to a header.
+    /// Enabled with `subtext` option.
+    Subtext,
 }
 
 /// Alignment of a single table cell.
@@ -231,7 +275,7 @@ pub struct NodeTable {
 }
 
 /// An inline [code span](https://github.github.com/gfm/#code-spans).
-#[derive(Debug, Clone, PartialEq, Eq)]
+#[derive(Default, Debug, Clone, PartialEq, Eq)]
 pub struct NodeCode {
     /// The number of backticks
     pub num_backticks: usize,
@@ -244,7 +288,7 @@ pub struct NodeCode {
 }
 
 /// The details of a link's destination, or an image's source.
-#[derive(Debug, Clone, PartialEq, Eq)]
+#[derive(Default, Debug, Clone, PartialEq, Eq)]
 pub struct NodeLink {
     /// The URL for the link destination or image source.
     pub url: String,
@@ -257,7 +301,7 @@ pub struct NodeLink {
 }
 
 /// The details of a wikilink's destination.
-#[derive(Debug, Clone, PartialEq, Eq)]
+#[derive(Default, Debug, Clone, PartialEq, Eq)]
 pub struct NodeWikiLink {
     /// The URL for the link destination.
     pub url: String,
@@ -287,6 +331,9 @@ pub struct NodeList {
     /// Whether the list is [tight](https://github.github.com/gfm/#tight), i.e. whether the
     /// paragraphs are wrapped in `<p>` tags when formatted as HTML.
     pub tight: bool,
+
+    /// Whether the list contains tasks (checkbox items)
+    pub is_task_list: bool,
 }
 
 /// The metadata of a description list
@@ -297,6 +344,10 @@ pub struct NodeDescriptionItem {
 
     /// Number of characters between the start of the list marker and the item text (including the list marker(s)).
     pub padding: usize,
+
+    /// Whether the list is [tight](https://github.github.com/gfm/#tight), i.e. whether the
+    /// paragraphs are wrapped in `<p>` tags when formatted as HTML.
+    pub tight: bool,
 }
 
 /// The type of list.
@@ -353,6 +404,10 @@ pub struct NodeCodeBlock {
     /// all, they are contained within this structure, rather than inserted into a child inline of
     /// any kind.
     pub literal: String,
+
+    /// Whether the code block was explicitly closed by a closing fence. This is only meaningful
+    /// for fenced code blocks; indented code blocks are treated as closed (true).
+    pub closed: bool,
 }
 
 /// The metadata of a heading.
@@ -363,6 +418,10 @@ pub struct NodeHeading {
 
     /// Whether the heading is setext (if not, ATX).
     pub setext: bool,
+
+    /// Whether this ATX heading had a closing sequence of trailing hashes.
+    /// Only meaningful for ATX headings (i.e. when `setext` is false).
+    pub closed: bool,
 }
 
 /// The metadata of an included HTML block.
@@ -399,6 +458,96 @@ pub struct NodeFootnoteReference {
     pub ix: u32,
 }
 
+/// The metadata of a multiline blockquote.
+#[derive(Default, Debug, Clone, Copy, PartialEq, Eq)]
+pub struct NodeMultilineBlockQuote {
+    /// The length of the fence.
+    pub fence_length: usize,
+
+    /// The indentation level of the fence marker.
+    pub fence_offset: usize,
+}
+
+/// An inline math span
+#[derive(Default, Debug, Clone, PartialEq, Eq)]
+pub struct NodeMath {
+    /// Whether this is dollar math (`$` or `$$`).
+    /// `false` indicates it is code math
+    pub dollar_math: bool,
+
+    /// Whether this is display math (using `$$`)
+    pub display_math: bool,
+
+    /// The literal contents of the math span.
+    /// As the contents are not interpreted as Markdown at all,
+    /// they are contained within this structure,
+    /// rather than inserted into a child inline of any kind.
+    pub literal: String,
+}
+
+/// The metadata of an Alert node.
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct NodeAlert {
+    /// Type of alert
+    pub alert_type: AlertType,
+
+    /// Overridden title. If `None`, then use the default title.
+    pub title: Option<String>,
+
+    /// Originated from a multiline blockquote.
+    pub multiline: bool,
+
+    /// The length of the fence (multiline only).
+    pub fence_length: usize,
+
+    /// The indentation level of the fence marker (multiline only)
+    pub fence_offset: usize,
+}
+
+/// The type of alert.
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Default)]
+pub enum AlertType {
+    /// Useful information that users should know, even when skimming content
+    #[default]
+    Note,
+
+    /// Helpful advice for doing things better or more easily
+    Tip,
+
+    /// Key information users need to know to achieve their goal
+    Important,
+
+    /// Urgent info that needs immediate user attention to avoid problems
+    Warning,
+
+    /// Advises about risks or negative outcomes of certain actions
+    Caution,
+}
+
+impl AlertType {
+    /// Returns the default title for an alert type
+    pub fn default_title(&self) -> String {
+        match *self {
+            AlertType::Note => String::from("Note"),
+            AlertType::Tip => String::from("Tip"),
+            AlertType::Important => String::from("Important"),
+            AlertType::Warning => String::from("Warning"),
+            AlertType::Caution => String::from("Caution"),
+        }
+    }
+
+    /// Returns the CSS class to use for an alert type
+    pub fn css_class(&self) -> String {
+        match *self {
+            AlertType::Note => String::from("markdown-alert-note"),
+            AlertType::Tip => String::from("markdown-alert-tip"),
+            AlertType::Important => String::from("markdown-alert-important"),
+            AlertType::Warning => String::from("markdown-alert-warning"),
+            AlertType::Caution => String::from("markdown-alert-caution"),
+        }
+    }
+}
+
 impl NodeValue {
     /// Indicates whether this node is a block node or inline node.
     pub fn block(&self) -> bool {
@@ -423,6 +572,8 @@ impl NodeValue {
                 | NodeValue::TableCell
                 | NodeValue::TaskItem(..)
                 | NodeValue::MultilineBlockQuote(_)
+                | NodeValue::Alert(_)
+                | NodeValue::Subtext
         )
     }
 
@@ -430,14 +581,17 @@ impl NodeValue {
     pub fn contains_inlines(&self) -> bool {
         matches!(
             *self,
-            NodeValue::Paragraph | NodeValue::Heading(..) | NodeValue::TableCell
+            NodeValue::Paragraph
+                | NodeValue::Heading(..)
+                | NodeValue::TableCell
+                | NodeValue::Subtext
         )
     }
 
     /// Return a reference to the text of a `Text` inline, if this node is one.
     ///
     /// Convenience method.
-    pub fn text(&self) -> Option<&String> {
+    pub fn text(&self) -> Option<&str> {
         match *self {
             NodeValue::Text(ref t) => Some(t),
             _ => None,
@@ -447,7 +601,7 @@ impl NodeValue {
     /// Return a mutable reference to the text of a `Text` inline, if this node is one.
     ///
     /// Convenience method.
-    pub fn text_mut(&mut self) -> Option<&mut String> {
+    pub fn text_mut(&mut self) -> Option<&mut Cow<'static, str>> {
         match *self {
             NodeValue::Text(ref mut t) => Some(t),
             _ => None,
@@ -457,7 +611,10 @@ impl NodeValue {
     pub(crate) fn accepts_lines(&self) -> bool {
         matches!(
             *self,
-            NodeValue::Paragraph | NodeValue::Heading(..) | NodeValue::CodeBlock(..)
+            NodeValue::Paragraph
+                | NodeValue::Heading(..)
+                | NodeValue::CodeBlock(..)
+                | NodeValue::Subtext
         )
     }
 
@@ -489,6 +646,7 @@ impl NodeValue {
             NodeValue::Strong => "strong",
             NodeValue::Code(..) => "code",
             NodeValue::HtmlInline(..) => "html_inline",
+            NodeValue::Raw(..) => "raw",
             NodeValue::Strikethrough => "strikethrough",
             NodeValue::FrontMatter(_) => "frontmatter",
             NodeValue::TaskItem { .. } => "taskitem",
@@ -500,6 +658,12 @@ impl NodeValue {
             NodeValue::Escaped => "escaped",
             NodeValue::Math(..) => "math",
             NodeValue::WikiLink(..) => "wikilink",
+            NodeValue::Underline => "underline",
+            NodeValue::Subscript => "subscript",
+            NodeValue::SpoileredText => "spoiler",
+            NodeValue::EscapedTag(_) => "escaped_tag",
+            NodeValue::Alert(_) => "alert",
+            NodeValue::Subtext => "subtext",
         }
     }
 }
@@ -515,14 +679,29 @@ pub struct Ast {
 
     /// The positions in the source document this node comes from.
     pub sourcepos: Sourcepos,
-    pub(crate) internal_offset: usize,
 
     pub(crate) content: String,
     pub(crate) open: bool,
     pub(crate) last_line_blank: bool,
     pub(crate) table_visited: bool,
+    pub(crate) line_offsets: Vec<usize>,
 }
 
+#[allow(dead_code)]
+#[cfg(target_pointer_width = "64")]
+/// Assert the size of Ast is 128 bytes. It's pretty big; let's stop it getting
+/// bigger.
+const AST_SIZE_ASSERTION: [u8; 128] = [0; std::mem::size_of::<Ast>()];
+
+#[allow(dead_code)]
+#[cfg(target_pointer_width = "64")]
+/// Assert the total size of what we allocate in the Arena, for reference.
+///
+/// Note that the size adds to Ast:
+/// * 8 bytes for RefCell.
+/// * 40 bytes for arena_tree::Node's 5 pointers.
+const AST_NODE_SIZE_ASSERTION: [u8; 176] = [0; std::mem::size_of::<AstNode<'_>>()];
+
 /// Represents the position in the source Markdown this node was rendered from.
 #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
 pub struct Sourcepos {
@@ -558,7 +737,7 @@ impl From<(usize, usize, usize, usize)> for Sourcepos {
 }
 
 /// Represents the 1-based line and column positions of a given character.
-#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
+#[derive(Default, Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
 pub struct LineColumn {
     /// The 1-based line number of the character.
     pub line: usize,
@@ -586,139 +765,280 @@ impl LineColumn {
 }
 
 impl Ast {
-    /// Create a new AST node with the given value.
+    /// Create a new AST node with the given value and starting sourcepos. The
+    /// end column is set to zero; it is expected this will be set manually
+    /// or later in the parse.  Use [`Ast::new_with_sourcepos`] if you have full
+    /// sourcepos.
     pub fn new(value: NodeValue, start: LineColumn) -> Self {
         Ast {
             value,
             content: String::new(),
             sourcepos: (start.line, start.column, start.line, 0).into(),
-            internal_offset: 0,
             open: true,
             last_line_blank: false,
             table_visited: false,
+            line_offsets: Vec::new(),
+        }
+    }
+
+    /// Create a new AST node with the given value and sourcepos.
+    pub fn new_with_sourcepos(value: NodeValue, sourcepos: Sourcepos) -> Self {
+        Ast {
+            value,
+            content: String::new(),
+            sourcepos,
+            open: true,
+            last_line_blank: false,
+            table_visited: false,
+            line_offsets: Vec::new(),
         }
     }
 }
 
 /// The type of a node within the document.
 ///
-/// It is bound by the lifetime `'a`, which corresponds to the `Arena` nodes are allocated in.
-/// Child `Ast`s are wrapped in `RefCell` for interior mutability.
-pub type AstNode<'a> = Node<'a, RefCell<Ast>>;
+/// It is bound by the lifetime `'a`, which corresponds to the `Arena` nodes are
+/// allocated in. Child `Ast`s are wrapped in `RefCell` for interior mutability.
+///
+/// You can construct a new `AstNode` from a `NodeValue` using the `From` trait:
+///
+/// ```rust
+/// # use comrak::nodes::{AstNode, NodeValue};
+/// let root = AstNode::from(NodeValue::Document);
+/// ```
+///
+/// Note that no sourcepos information is given to the created node. If you wish
+/// to assign sourcepos information, use the `From` trait to create an `AstNode`
+/// from an `Ast`:
+///
+/// ```rust
+/// # use comrak::nodes::{Ast, AstNode, NodeValue};
+/// let root = AstNode::from(Ast::new_with_sourcepos(
+///     NodeValue::Paragraph,
+///     (4, 1, 4, 10).into(),
+/// ));
+/// ```
+///
+/// For practical use, you'll probably need it allocated in an `Arena`, in which
+/// case you can use `.into()` to simplify creation:
+///
+/// ```rust
+/// # use comrak::{nodes::{AstNode, NodeValue}, Arena};
+/// # let arena = Arena::<AstNode>::new();
+/// let node_in_arena = arena.alloc(NodeValue::Document.into());
+/// ```
+pub type AstNode<'a> = arena_tree::Node<'a, RefCell<Ast>>;
 
-pub(crate) fn last_child_is_open<'a>(node: &'a AstNode<'a>) -> bool {
-    node.last_child().map_or(false, |n| n.data.borrow().open)
-}
+/// A reference to a node in an arena.  Unless you are manually creating nodes
+/// before the arena, this is the type you will see most often.
+pub type Node<'a> = &'a AstNode<'a>;
 
-/// Returns true if the given node can contain a node with the given value.
-pub fn can_contain_type<'a>(node: &'a AstNode<'a>, child: &NodeValue) -> bool {
-    match *child {
-        NodeValue::Document => {
-            return false;
-        }
-        NodeValue::FrontMatter(_) => {
-            return matches!(node.data.borrow().value, NodeValue::Document);
-        }
-        _ => {}
-    }
-
-    match node.data.borrow().value {
-        NodeValue::Document
-        | NodeValue::BlockQuote
-        | NodeValue::FootnoteDefinition(_)
-        | NodeValue::DescriptionTerm
-        | NodeValue::DescriptionDetails
-        | NodeValue::Item(..)
-        | NodeValue::TaskItem(..) => {
-            child.block() && !matches!(*child, NodeValue::Item(..) | NodeValue::TaskItem(..))
-        }
-
-        NodeValue::List(..) => matches!(*child, NodeValue::Item(..) | NodeValue::TaskItem(..)),
-
-        NodeValue::DescriptionList => matches!(*child, NodeValue::DescriptionItem(_)),
-
-        NodeValue::DescriptionItem(_) => matches!(
-            *child,
-            NodeValue::DescriptionTerm | NodeValue::DescriptionDetails
-        ),
-
-        #[cfg(feature = "shortcodes")]
-        NodeValue::ShortCode(..) => !child.block(),
-
-        NodeValue::Paragraph
-        | NodeValue::Heading(..)
-        | NodeValue::Emph
-        | NodeValue::Strong
-        | NodeValue::Link(..)
-        | NodeValue::Image(..)
-        | NodeValue::WikiLink(..) => !child.block(),
-
-        NodeValue::Table(..) => matches!(*child, NodeValue::TableRow(..)),
-
-        NodeValue::TableRow(..) => matches!(*child, NodeValue::TableCell),
-
-        #[cfg(not(feature = "shortcodes"))]
-        NodeValue::TableCell => matches!(
-            *child,
-            NodeValue::Text(..)
-                | NodeValue::Code(..)
-                | NodeValue::Emph
-                | NodeValue::Strong
-                | NodeValue::Link(..)
-                | NodeValue::Image(..)
-                | NodeValue::Strikethrough
-                | NodeValue::HtmlInline(..)
-                | NodeValue::Math(..)
-                | NodeValue::WikiLink(..)
-        ),
-
-        #[cfg(feature = "shortcodes")]
-        NodeValue::TableCell => matches!(
-            *child,
-            NodeValue::Text(..)
-                | NodeValue::Code(..)
-                | NodeValue::Emph
-                | NodeValue::Strong
-                | NodeValue::Link(..)
-                | NodeValue::Image(..)
-                | NodeValue::ShortCode(..)
-                | NodeValue::Strikethrough
-                | NodeValue::HtmlInline(..)
-                | NodeValue::Math(..)
-                | NodeValue::WikiLink(..)
-        ),
-
-        NodeValue::MultilineBlockQuote(_) => {
-            child.block() && !matches!(*child, NodeValue::Item(..) | NodeValue::TaskItem(..))
-        }
-
-        _ => false,
+impl<'a> From<NodeValue> for AstNode<'a> {
+    /// Create a new AST node with the given value. The sourcepos is set to (0,0)-(0,0).
+    fn from(value: NodeValue) -> Self {
+        arena_tree::Node::new(RefCell::new(Ast::new(value, LineColumn::default())))
     }
 }
 
-pub(crate) fn ends_with_blank_line<'a>(node: &'a AstNode<'a>) -> bool {
-    let mut it = Some(node);
-    while let Some(cur) = it {
-        if cur.data.borrow().last_line_blank {
-            return true;
-        }
-        match cur.data.borrow().value {
-            NodeValue::List(..) | NodeValue::Item(..) | NodeValue::TaskItem(..) => {
-                it = cur.last_child()
+impl<'a> From<Ast> for AstNode<'a> {
+    /// Create a new AST node with the given Ast.
+    fn from(ast: Ast) -> Self {
+        arena_tree::Node::new(RefCell::new(ast))
+    }
+}
+
+/// Validation errors produced by [arena_tree::Node::validate].
+#[derive(Debug, Clone)]
+pub enum ValidationError<'a> {
+    /// The type of a child node is not allowed in the parent node. This can happen when an inline
+    /// node is found in a block container, a block is found in an inline node, etc.
+    InvalidChildType {
+        /// The parent node.
+        parent: Node<'a>,
+        /// The child node.
+        child: Node<'a>,
+    },
+}
+
+impl<'a> arena_tree::Node<'a, RefCell<Ast>> {
+    /// Returns true if the given node can contain a node with the given value.
+    pub fn can_contain_type(&self, child: &NodeValue) -> bool {
+        match *child {
+            NodeValue::Document => {
+                return false;
             }
-            _ => it = None,
-        };
-    }
-    false
-}
+            NodeValue::FrontMatter(_) => {
+                return node_matches!(self, NodeValue::Document);
+            }
+            _ => {}
+        }
 
-pub(crate) fn containing_block<'a>(node: &'a AstNode<'a>) -> Option<&'a AstNode<'a>> {
-    let mut ch = Some(node);
-    while let Some(n) = ch {
-        if n.data.borrow().value.block() {
-            return Some(n);
+        match self.data().value {
+            NodeValue::Document
+            | NodeValue::BlockQuote
+            | NodeValue::FootnoteDefinition(_)
+            | NodeValue::DescriptionTerm
+            | NodeValue::DescriptionDetails
+            | NodeValue::Item(..)
+            | NodeValue::TaskItem(..) => {
+                child.block() && !matches!(*child, NodeValue::Item(..) | NodeValue::TaskItem(..))
+            }
+            NodeValue::List(..) => matches!(*child, NodeValue::Item(..) | NodeValue::TaskItem(..)),
+            NodeValue::DescriptionList => matches!(*child, NodeValue::DescriptionItem(_)),
+            NodeValue::DescriptionItem(_) => matches!(
+                *child,
+                NodeValue::DescriptionTerm | NodeValue::DescriptionDetails
+            ),
+            #[cfg(feature = "shortcodes")]
+            NodeValue::ShortCode(..) => !child.block(),
+            NodeValue::Paragraph
+            | NodeValue::Heading(..)
+            | NodeValue::Emph
+            | NodeValue::Strong
+            | NodeValue::Link(..)
+            | NodeValue::Image(..)
+            | NodeValue::WikiLink(..)
+            | NodeValue::Strikethrough
+            | NodeValue::Superscript
+            | NodeValue::SpoileredText
+            | NodeValue::Underline
+            | NodeValue::Subscript
+            | NodeValue::Subtext
+            // XXX: this is quite a hack: the EscapedTag _contains_ whatever was
+            // possibly going to fall into the spoiler. This should be fixed in
+            // inlines.
+            | NodeValue::EscapedTag(_)
+            => !child.block(),
+            NodeValue::Table(..) => matches!(*child, NodeValue::TableRow(..)),
+            NodeValue::TableRow(..) => matches!(*child, NodeValue::TableCell),
+            #[cfg(not(feature = "shortcodes"))]
+            NodeValue::TableCell => matches!(
+                *child,
+                NodeValue::Text(..)
+                    | NodeValue::Code(..)
+                    | NodeValue::Emph
+                    | NodeValue::Strong
+                    | NodeValue::Link(..)
+                    | NodeValue::Image(..)
+                    | NodeValue::Strikethrough
+                    | NodeValue::HtmlInline(..)
+                    | NodeValue::Math(..)
+                    | NodeValue::WikiLink(..)
+                    | NodeValue::FootnoteReference(..)
+                    | NodeValue::Superscript
+                    | NodeValue::SpoileredText
+                    | NodeValue::Underline
+                    | NodeValue::Subscript
+                    | NodeValue::TaskItem(_)
+                    | NodeValue::Escaped
+                    | NodeValue::EscapedTag(_)
+            ),
+            #[cfg(feature = "shortcodes")]
+            NodeValue::TableCell => matches!(
+                *child,
+                NodeValue::Text(..)
+                | NodeValue::Code(..)
+                | NodeValue::Emph
+                | NodeValue::Strong
+                | NodeValue::Link(..)
+                | NodeValue::Image(..)
+                | NodeValue::Strikethrough
+                | NodeValue::HtmlInline(..)
+                | NodeValue::Math(..)
+                | NodeValue::WikiLink(..)
+                | NodeValue::FootnoteReference(..)
+                | NodeValue::Superscript
+                | NodeValue::SpoileredText
+                | NodeValue::Underline
+                | NodeValue::Subscript
+                | NodeValue::ShortCode(..)
+                | NodeValue::TaskItem(_)
+                | NodeValue::Escaped
+                | NodeValue::EscapedTag(_)
+            ),
+            NodeValue::MultilineBlockQuote(_) => {
+                child.block() && !matches!(*child, NodeValue::Item(..) | NodeValue::TaskItem(..))
+            }
+            NodeValue::Alert(_) => {
+                child.block() && !matches!(*child, NodeValue::Item(..) | NodeValue::TaskItem(..))
+            }
+
+            NodeValue::Escaped => matches!(*child, NodeValue::Text(_)),
+
+            NodeValue::FrontMatter(_)
+            | NodeValue::CodeBlock(_)
+            | NodeValue::HtmlBlock(_)
+            | NodeValue::ThematicBreak
+            | NodeValue::Text(_)
+            | NodeValue::SoftBreak
+            | NodeValue::LineBreak
+            | NodeValue::Code(_)
+            | NodeValue::HtmlInline(_)
+            | NodeValue::Raw(_)
+            | NodeValue::FootnoteReference(_)
+            | NodeValue::Math(_) => false,
         }
-        ch = n.parent();
     }
-    None
+
+    /// The Comrak representation of a markdown node in Rust isn't strict enough to rule out
+    /// invalid trees according to the CommonMark specification. One simple example is that block
+    /// containers, such as lists, should only contain blocks, but it's possible to put naked
+    /// inline text in a list item. Such invalid trees can lead Comrak to generate incorrect output
+    /// if rendered.
+    ///
+    /// This method performs additional structural checks to ensure that a markdown AST is valid
+    /// according to the CommonMark specification.
+    ///
+    /// Note that those invalid trees can only be generated programmatically. Parsing markdown with
+    /// Comrak, on the other hand, should always produce a valid tree.
+    pub fn validate(&'a self) -> Result<(), ValidationError<'a>> {
+        let mut stack = vec![self];
+
+        while let Some(node) = stack.pop() {
+            // Check that this node type is valid wrt to the type of its parent.
+            if let Some(parent) = node.parent() {
+                if !parent.can_contain_type(&node.data().value) {
+                    return Err(ValidationError::InvalidChildType {
+                        parent,
+                        child: node,
+                    });
+                }
+            }
+
+            stack.extend(node.children());
+        }
+
+        Ok(())
+    }
+
+    pub(crate) fn last_child_is_open(&self) -> bool {
+        self.last_child().map_or(false, |n| n.data().open)
+    }
+
+    pub(crate) fn ends_with_blank_line(&self) -> bool {
+        let mut it = Some(self);
+        while let Some(cur) = it {
+            if cur.data().last_line_blank {
+                return true;
+            }
+            match cur.data().value {
+                NodeValue::List(..) | NodeValue::Item(..) | NodeValue::TaskItem(..) => {
+                    it = cur.last_child()
+                }
+                _ => it = None,
+            };
+        }
+        false
+    }
+
+    pub(crate) fn containing_block(&'a self) -> Option<Node<'a>> {
+        let mut ch = Some(self);
+        while let Some(n) = ch {
+            if n.data().value.block() {
+                return Some(n);
+            }
+            ch = n.parent();
+        }
+        None
+    }
 }
