@@ -11,8 +11,8 @@ pub use camino::{Utf8Path as Path, Utf8PathBuf as PathBuf};
 
 pub use error::{CacheError, Error, HttpError, InvalidUrl, InvalidUrlError};
 pub use index::{
-    git::CRATES_IO_INDEX, sparse::CRATES_IO_HTTP_INDEX, GitIndex, IndexCache, IndexLocation,
-    IndexPath, IndexUrl, SparseIndex,
+    GitIndex, IndexCache, IndexLocation, IndexPath, IndexUrl, SparseIndex, git::CRATES_IO_INDEX,
+    sparse::CRATES_IO_HTTP_INDEX,
 };
 pub use krate::{IndexDependency, IndexKrate, IndexVersion};
 pub use krate_name::KrateName;
